
# This is the "large stone" level operation of the bldeif infrastructure,
# wherein the two connections held by the connector get instantiated, 
# connected to their respective systems and exercised.

import sys, os, platform
import time
import re
from collections import OrderedDict

from bldeif.utils.eif_exception   import FatalError, ConfigurationError, OperationalError
from bldeif.utils.claslo          import ClassLoader
from bldeif.utils.chronuti        import TimeStamp

##############################################################################################

__version__ = "1.1.3"

PLUGIN_SPEC_PATTERN       = re.compile(r'^(?P<plugin_class>\w+)\s*\((?P<plugin_config>[^\)]*)\)\s*$')
PLAIN_PLUGIN_SPEC_PATTERN = re.compile(r'(?P<plugin_class>\w+)\s*$')

ARCHITECTURE_PKG = 'bldeif'

##############################################################################################

class BLDConnector:

    def __init__(self, config, logger):
        self.config = config
        self.log    = logger

        conn_sections = []
        try:
            conn_sections = [section for section in config.topLevels()
                                      if section not in ['Service']]
            if len(conn_sections) != 2:
                raise ConfigurationError('Config does not contain two connection sections')
        except Exception:
            raise ConfigurationError('Config file lacks sufficient information for BLDConnector')

        self.agicen_conn = None
        self.bld_conn    = None

        self.bld_name   = [name for name in conn_sections if not name.startswith('AgileCentral')][0]
        self.log.info("Agile Central BLD Connector for %s, version %s" % (self.bld_name, __version__))
        uname_fields = platform.uname()
        self.log.info("Python platform: %s %s" % (uname_fields[0], uname_fields[2]))
        self.log.info("Python  version: %s" % sys.version.split()[0])

        self.internalizeConfig(config)

        droid = ClassLoader()
        agicen_conn_class_name = config.connectionClassName('AgileCentral')
        bld_conn_class_name    = config.connectionClassName(self.bld_name)

        try:
            self.agicen_conn_class = droid.loadConnectionClass(agicen_conn_class_name, pkgdir=ARCHITECTURE_PKG)
        except Exception as msg:
            raise FatalError('Unable to load AgileCentralBLDConnection class, %s' % msg)
        try:
            self.bld_conn_class = droid.loadConnectionClass(bld_conn_class_name, pkgdir=ARCHITECTURE_PKG)
        except Exception as msg:
            raise FatalError('Unable to load %sConnection class, %s' % (self.bld_name, msg))

        self.establishConnections()

        self.setupProjectScope()

        if not self.validate():  # basically just calls validate on both connection instances
            raise ConfigurationError("Validation failed")
        self.log.info("Initialization complete: Delegate connections operational, ready for scan/reflect ops")


    def internalizeConfig(self, config):
        self.agicen_conf = config.topLevel('AgileCentral')
        self.bld_conf    = config.topLevel(self.bld_name)
        if not 'AgileCentral_DefaultBuildProject' in self.bld_conf:
            msg = "The %s section of the config is missing AgileCentral_DefaultBuildProject property"
            raise ConfigurationError(msg % self.bld_name)
        if not self.bld_conf['AgileCentral_DefaultBuildProject']:  # but no value exists for this...
            msg = "The %s section of the config is missing a value for AgileCentral_DefaultBuildProject property"
            raise ConfigurationError(msg % self.bld_name)
        self.agicen_conf['Project'] = self.bld_conf['AgileCentral_DefaultBuildProject']
        self.svc_conf    = config.topLevel('Service')
        self.max_builds  = self.svc_conf.get('MaxBuilds', 20)

        valid_config_items = ['Preview', 'LogLevel', 'MaxBuilds', 'ShowVCSData', 'SecurityLevel' ]
        svc_conf = config.topLevel('Service')
        invalid_config_items = [item for item in svc_conf.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "Service section of the config contained these invalid entries: %s" % ", ".join(invalid_config_items)
            raise ConfigurationError(problem)


    def establishConnections(self):
        self.agicen_conn = self.agicen_conn_class(self.agicen_conf, self.log)
        self.bld_conn    =    self.bld_conn_class(self.bld_conf,    self.log)

        self.bld_conn.connect()  # we do this before agicen_conn to be able to get the bld backend version
        bld_backend_version = self.bld_conn.getBackendVersion()

        if self.agicen_conn and self.bld_conn and getattr(self.agicen_conn, 'setIntegrationHeader'):
            agicen_headers = {'name'    : 'Agile Central BLDConnector for %s' % self.bld_name,
                              'version' : __version__,
                              'vendor'  : 'CA Technologies',
                              'other_version' : bld_backend_version
                            }
            self.agicen_conn.setIntegrationHeader(agicen_headers)
        self.agicen_conn.setSourceIdentification(self.bld_conn.name(), self.bld_conn.backend_version)
        self.agicen_conn.connect()  # so we can use it in our X-Rally-Integrations header items here


    def setupProjectScope(self):
        """
            create a list of AgileCentral_Project values, start with the default project value
            and then add add as you see overrides in the config.
            eventually, we'll strip out any duplicates
        """
        default_project = self.agicen_conf['Project']
        ac_project_key = 'AgileCentral_Project'
        self.target_projects = self.bld_conn._getTargetProjects(self.bld_conf, ac_project_key, default_project)


    def validate(self):
        """
            This calls the validate method on both the Agile Central and the BLD connections
        """
        self.log.info("Connector validation starting")

        if not self.agicen_conn.validate(self.target_projects):
            self.log.info("AgileCentralConnection validation failed")
            return False
        # if not self.agicen_conn.validateProjects(self.target_projects):
        #     self.log.info("AgileCentralConnection validation for Projects failed")
        #     return False
        self.log.info("AgileCentralConnection validation succeeded")

        if not self.bld_conn.validate():
            self.log.info("%sConnection validation failed" % self.bld_name)
            return False
        self.log.info("%sConnection validation succeeded" % self.bld_name)
        #self.bld_conn.dumpTargets()
        #self.bld_conn.showQualifiedActivities()    # whether it be Jobs or Plans or ...

        self.log.info("Connector validation completed")

        return True

    def targetActivityUnit(self):
        """
            Delegate responsibility for getting the target_activity_unit to the "foreign" spoke
        """
        return self.bld_conn.targetActivityUnit()


    def run(self, last_run_ts, extension):
        """
            The real beef is in the call to reflectBuildsInAgileCentral.
            The facility for extensions is not yet implemented for BLD connectors,
            so the pre and post batch calls are currently no-ops.
        """
        self.preBatch(extension)
        status, builds = self.reflectBuildsInAgileCentral(last_run_ts)
        self.postBatch(extension, status, builds)
        return status, builds


    def preBatch(self, extension):
        """
        """
        if extension and 'PreBatchAction' in extension:
            preba = extension['PreBatchAction']
            preba.service()


    def postBatch(self, extension, status, builds):
        """
        """
        if extension and 'PostBatchAction' in extension:
            postba = extension['PostBatchAction']
            postba.service(status, builds)


    def reflectBuildsInAgileCentral(self, last_run_ts):
        """
            The last run time is passed to Connection objects in UTC;
            they are responsible for converting if necessary. 
            Time in log messages is always reported in UTC (aka Z or Zulu time).
        """
        status = False
        agicen = self.agicen_conn
        bld    = self.bld_conn

        preview_mode = self.svc_conf.get('Preview', False)
        get_vcs_data = self.svc_conf.get('ShowVCSData', False)

        if preview_mode:
            self.log.info('***** Preview Mode *****   (no Builds will be created in Agile Central)')

        agicen_ref_ts, bld_ref_ts = self.getRefTimes(last_run_ts)
        recent_agicen_builds = agicen.getRecentBuilds(agicen_ref_ts, self.target_projects)
        recent_bld_builds    =    bld.getRecentBuilds(bld_ref_ts, get_vcs_data=get_vcs_data)
        unrecorded_builds = self.bld_conn._identifyUnrecordedBuilds(recent_agicen_builds, recent_bld_builds)
        self.log.info("unrecorded Builds count: %d" % len(unrecorded_builds))
        self.log.info("no more than %d builds per %s will be recorded on this run" % (self.max_builds, self.targetActivityUnit()))
        if get_vcs_data:
            self.dumpChangesetInfo(unrecorded_builds)

        recorded_builds = OrderedDict()
        builds_posted = {}
        # sort the unrecorded_builds into build chrono order, oldest to most recent, then project and job
        unrecorded_builds.sort(key=lambda build_info: (build_info[1].completed_timestamp.asEpochSeconds(), build_info[2], build_info[1]))
        self.log.debug("About to process %d unrecorded builds" % len(unrecorded_builds))

        for item in unrecorded_builds:
            # while item may have more elements, we only need the first 3 of them
            activity, build, project = item[:3]

            if not build.done():
                self.log.warn("%s #%s activity/build was not processed because is still running" % (activity, build.number))
                continue

            #self.log.debug("current activity: %s  build: %s" % (activity, build))
            if not activity in builds_posted:
                builds_posted[activity] = 0
            if builds_posted[activity] >= self.max_builds:
                continue
            if preview_mode:
                continue

            activity_name, activity_path, activity_uri = activity.getActivityIdentifiers()
            activity_label = activity_path or activity_name

            try:
                changesets, build_definition = agicen.prepAgileCentralBuildPrerequisites(activity_label, activity_uri, build, project)
            except Exception as msg:
                self.log.error('OperationalException prepACBuildPrerequisites - %s' % msg)
                continue

            try:
                agicen_build, status = self.postBuildToAgileCentral(build_definition, build, changesets, activity)
            except Exception as msg:
                self.log.error('OperationalException postingACBuild - %s' % msg)
                continue

            if agicen_build and status == 'posted':
                builds_posted[activity] += 1
                if activity not in recorded_builds:
                    recorded_builds[activity] = []
                recorded_builds[activity].append(agicen_build)
            status = True

        return status, recorded_builds


    def postBuildToAgileCentral(self, build_defn, build, changesets, activity):
        """
            add the changesets incoming parameter as the Build's Changesets collection
            and post the Build to AC
        """
        desc = '%s #%s | %s | %s  not yet reflected in Agile Central'
        bts = build.completed_timestamp.asISOString()
        self.log.debug(desc % (build_defn.Name, build.number, build.result, bts))
        build_data = build.as_tuple_data()
        info = OrderedDict(build_data)
        info['BuildDefinition'] = build_defn
        if changesets:
            info['Changesets'] = changesets
        existing_agicen_build = self.agicen_conn.buildExists(build_defn, build.number)
        if existing_agicen_build:
            self.log.debug('Build #{0} for {1} already recorded, skipping...'.format(build.number, activity))
            return existing_agicen_build, 'skipped'
        agicen_build = self.agicen_conn.createBuild(info)
        return agicen_build, 'posted'


    def getRefTimes(self, last_run_ts):
        """
            last_run_ts is provided as a TimeStamp instance.
            Return a two-tuple of the reference TimeStamp instance to be used for obtaining the
            recent Builds in AgileCentral and the reference TimeStamp instance to be used for
            obtaining the recent builds in the target BLD system.
        """
        agicen_ref_ts = last_run_ts.minus(self.agicen_conn.lookback, units='seconds')
        bld_ref_ts    = last_run_ts.minus(self.bld_conn.lookback,    units='seconds')
        return agicen_ref_ts, bld_ref_ts


    def _showBuildInformation(self, agicen_builds, bld_builds):
        for project, activity_builds in agicen_builds.items():
            print("Agile Central project: %s" % project)
            for activity, builds in activity_builds.items():
                print("    %-36.36s : %3d build items" % (activity, len(builds)))
        print("")


    def dumpChangesetInfo(self, builds):
        """
            in Jenkins JenkinsJob, JenkinsBuild, 'RallyProject1', 'jenkins view'
            in Bamboo  BambooPlan, BambooBuild, 'RallyProject1'
        """
        for activity, build, *other in builds: # e.g. in Jenkins JenkinsJob, JenkinsBuild, 'Static', 'immovable wombat'
            if not build.changesets:
                continue
            self.log.debug(build)
            for cs in build.changesets:
                self.log.debug(cs)


    def detectCommitsForJenkinsBuild(self, build):
        shas = set([cs.id for cs in build.changesets])

        bacs = []
        for sha in shas:
            ac_changeset = self.agicen_conn.retrieveChangeset(sha)
            if ac_changeset:
                bacs.append(ac_changeset)

        return bacs

####################################################################################
